"""Test filenames with human parsed correct results."""

TEST_COMIC_FIELDS = {
    "series": "Long Series Name",
    "issue": "001",
    "year": "2000",
    "original_format": "TPB",
    "scan_info": "Releaser",
    "title": "Title",
    "ext": "cbz",
}
TEST_COMIC_FIELDS_VOL = {
    "series": "Long Series Name",
    "volume": "1",
    "issue": "001",
    "year": "2000",
    "title": "Title",
    "original_format": "TPB",
    "scan_info": "Releaser",
    "ext": "cbz",
}
TEST_COMIC_VOL_ONLY = {
    "series": "Long Series Name",
    "volume": "1",
    "title": "Title",
    "original_format": "TPB",
    "year": "2000",
    "scan_info": "Releaser & Releaser-Releaser",
    "ext": "cbr",
}

FNS = {
    "Night of 1000 Wolves 001 (2013).cbz": {
        "series": "Night of 1000 Wolves",
        "issue": "001",
        "year": "2013",
        "ext": "cbz",
    },
    "19687 Sandman 53.cbz": {
        "series": "19687 Sandman",
        "issue": "53",
        "ext": "cbz",
    },
    "33475 OMAC v3 2.cbr": {
        "series": "33475 OMAC",
        "volume": "3",
        "issue": "2",
        "ext": "cbr",
    },
    "Long Series Name 001 (2000) Title (TPB) (Releaser).cbz": TEST_COMIC_FIELDS,
    "Long Series Name #001 (2000) Title (TPB) (Releaser).cbz": TEST_COMIC_FIELDS,
    "Long Series Name (2000) 001 Title (TPB) (Releaser).cbz": TEST_COMIC_FIELDS,
    "Long Series Name (2000) #001 Title (TPB) (Releaser).cbz": TEST_COMIC_FIELDS,
    "Long Series Name v1 (2000) #001 "
    "Title (TPB) (Releaser).cbz": TEST_COMIC_FIELDS_VOL,
    "Long Series Name 001 (2000) (TPB-Releaser) Title.cbz": TEST_COMIC_FIELDS,
    "Long Series Name Vol 1 "
    "(2000) (TPB) (Releaser & Releaser-Releaser) Title.cbr": TEST_COMIC_VOL_ONLY,
    "Ultimate Craziness (2019) (Digital) (Friends-of-Bill).cbr": {
        "series": "Ultimate Craziness",
        "year": "2019",
        "ext": "cbr",
        "original_format": "Digital",
        "scan_info": "Friends-of-Bill",
    },
    "Jimmy Stocks Love Chain (2005) (digital) (The Magicians-Empire).cbr": {
        "series": "Jimmy Stocks Love Chain",
        "year": "2005",
        "ext": "cbr",
        "original_format": "digital",
        "scan_info": "The Magicians-Empire",
    },
    "Arkenstone Vol. 01 - The Smell of Burnt Toast (2020) (digital) (My-brother).cbr": {
        "series": "Arkenstone",
        "volume": "01",
        "year": "2020",
        "ext": "cbr",
        "scan_info": "My-brother",
        "title": "The Smell of Burnt Toast",
        "original_format": "digital",
    },
    "Bardude - The Last Thing I Remember.cbz": {
        "series": "Bardude",
        "title": "The Last Thing I Remember",
        "ext": "cbz",
    },
    "Drunkguy - The Man Without Fear - 01.cbz": {
        "series": "Drunkguy",
        "title": "The Man Without Fear",
        "issue": "01",
        "ext": "cbz",
    },
    "The_Arkenstone_v03_(2002)_(Digital)_(DR_&amp;_Quenya-Elves).cbr": {
        "series": "The Arkenstone",
        "volume": "03",
        "year": "2002",
        "ext": "cbr",
        "scan_info": "DR &amp; Quenya-Elves",
        "original_format": "Digital",
    },
    "Kartalk v01 001 - Fear the Brakes (2004) (digital) (Son of Ultron-EMpire).cbr": {
        "series": "Kartalk",
        "volume": "01",
        "issue": "001",
        "year": "2004",
        "ext": "cbr",
        "original_format": "digital",
        "scan_info": "Son of Ultron-EMpire",
        "title": "Fear the Brakes",
    },
    "Kartalk Library Edition v01 (1992) (digital) (Son of Ultron-Empire).cbr": {
        "series": "Kartalk Library Edition",
        "volume": "01",
        "year": "1992",
        "ext": "cbr",
        "original_format": "digital",
        "scan_info": "Son of Ultron-Empire",
    },
    "Kind of Deadly v02 - Last Bullet (2006) (Digital) (Zone-Empire).cbr": {
        "series": "Kind of Deadly",
        "volume": "02",
        "year": "2006",
        "ext": "cbr",
        "original_format": "Digital",
        "scan_info": "Zone-Empire",
        "title": "Last Bullet",
    },
    "Jeremy John - A Big Long Title (2017) (digital-Minutement).cbz": {
        "series": "Jeremy John",
        "title": "A Big Long Title",
        "year": "2017",
        "ext": "cbz",
        "original_format": "digital",
        "scan_info": "Minutement",
    },
    "Jeremy John 001 (2006) (digital (Minutemen-Faessla).cbz": {
        "series": "Jeremy John",
        "issue": "001",
        "year": "2006",
        "ext": "cbz",
        "scan_info": "Minutemen-Faessla",
        # "original_format": "digital",
        "remainders": ("(digital",),
    },
    "Jeremy John 003 (2007) (4 covers) (digital) (Minutemen-Faessla).cbz": {
        "series": "Jeremy John",
        "issue": "003",
        "year": "2007",
        "ext": "cbz",
        "scan_info": "Minutemen-Faessla",
        "original_format": "digital",
        "remainders": ("(4 covers)",),
    },
    "Jeremy John v01 - Uninterested! (2007) (Digital) (Asgard-Empire).cbr": {
        "series": "Jeremy John",
        "volume": "01",
        "year": "2007",
        "ext": "cbr",
        "original_format": "Digital",
        "scan_info": "Asgard-Empire",
        "title": "Uninterested!",
    },
    "King of Skittles 01 (of 05) (2020) (digital) (Son of Ultron-Empire).cbr": {
        "series": "King of Skittles",
        "issue": "01",
        "issue_count": "05",
        "year": "2020",
        "ext": "cbr",
        "original_format": "digital",
        "scan_info": "Son of Ultron-Empire",
    },
    "Darkwad 011 (2019) (Digital) (Zone-Empire).cbr": {
        "series": "Darkwad",
        "issue": "011",
        "year": "2019",
        "ext": "cbr",
        "original_format": "Digital",
        "scan_info": "Zone-Empire",
    },
    "Darkwad by Carlos Zemo v01 - Knuckle Fight (2009) (Digital) (Zone-Empire).cbr": {
        "series": "Darkwad by Carlos Zemo",
        "volume": "01",
        "year": "2009",
        "ext": "cbr",
        "title": "Knuckle Fight",
        "original_format": "Digital",
        "scan_info": "Zone-Empire",
    },
    "The Walking Dead #002 (2003).cbz": {
        "series": "The Walking Dead",
        "issue": "002",
        "year": "2003",
        "ext": "cbz",
    },
    "The Walking Dead #3.cbz": {
        "series": "The Walking Dead",
        "issue": "3",
        "ext": "cbz",
    },
    "The Walking Dead 4.cbz": {
        "series": "The Walking Dead",
        "issue": "4",
        "ext": "cbz",
    },
    "A Fractional Comic 1.1.cbz": {
        "series": "A Fractional Comic",
        "issue": "1.1",
        "ext": "cbz",
    },
    "A Fractional Comic 8.54.cbz": {
        "series": "A Fractional Comic",
        "issue": "8.54",
        "ext": "cbz",
    },
    "Earth X #½.cbz": {
        "series": "Earth X",
        "issue": "½",
        "ext": "cbz",
    },
    "Avengers #001½.cbz": {
        "series": "Avengers",
        "issue": "001½",
        "ext": "cbz",
    },
    "The Amazing Spider-Man #78.BEY.cbz": {
        "series": "The Amazing Spider-Man",
        "issue": "78.BEY",
        "ext": "cbz",
    },
    "The Amazing Spider-Man #54.LR.cbz": {
        "series": "The Amazing Spider-Man",
        "issue": "54.LR",
        "ext": "cbz",
    },
    "Wolverine & the X-Men #27AU.cbz": {
        "series": "Wolverine & the X-Men",
        "issue": "27AU",
        "ext": "cbz",
    },
    "Fantastic Four #5AU.cbz": {
        "series": "Fantastic Four",
        "issue": "5AU",
        "ext": "cbz",
    },
}



"""Parse comic book archive names using the simple 'parse' parser."""
import re
from pathlib import Path
from typing import Union

from comicfn2dict.regex import (
    DASH_SPLIT_RE,
    EXTRA_SPACES_RE,
    ISSUE_ANYWHERE_RE,
    ISSUE_BEGIN_RE,
    ISSUE_COUNT_RE,
    ISSUE_END_RE,
    ISSUE_NUMBER_RE,
    ISSUE_TOKEN_RE,
    NON_SPACE_DIVIDER_RE,
    ORIGINAL_FORMAT_RE,
    ORIGINAL_FORMAT_SCAN_INFO_RE,
    REMAINING_GROUP_RE,
    SCAN_INFO_RE,
    VOLUME_RE,
    YEAR_BEGIN_RE,
    YEAR_END_RE,
    YEAR_TOKEN_RE,
)

_REMAINING_GROUP_KEYS = ("series", "title")


def _parse_ext(name, suffix, metadata):
    """Pop the extension from the pathname."""
    data = name.removesuffix(suffix)
    ext = suffix.lstrip(".")
    if ext:
        metadata["ext"] = ext
    return data


def _clean_dividers(data):
    """Replace non space dividers and clean extra spaces out of string."""
    data = NON_SPACE_DIVIDER_RE.sub(" ", data)
    return EXTRA_SPACES_RE.sub(" ", data)


def _get_data_list(path, metadata):
    """Prepare data list from a path or string."""
    if isinstance(path, str):
        path = path.strip()
    path = Path(path)
    data = _parse_ext(path.name, path.suffix, metadata)
    data = _clean_dividers(data)
    return DASH_SPLIT_RE.split(data)


def _paren_strip(value: str):
    """Strip spaces and parens."""
    return value.strip().strip("()").strip()


def _splicey_dicey(data_list, index, match, match_group: Union[int, str] = 0):
    """Replace a string token from a list with two strings and the value removed.

    And return the value.
    """
    value = match.group(match_group)
    data = data_list.pop(index)
    data_ends = []
    if data_before := data[: match.start()].strip():
        data_ends.append(data_before)
    if data_after := data[match.end() :].strip():
        data_ends.append(data_after)
    data_list[index:index] = data_ends
    return _paren_strip(value)


def _parse_original_format_and_scan_info(data_list, metadata):
    """Parse (ORIGINAL_FORMAT-SCAN_INFO)."""
    original_format = None
    scan_info = None
    index = 0
    match = None
    for data in data_list:
        match = ORIGINAL_FORMAT_SCAN_INFO_RE.search(data)
        if match:
            original_format = match.group("original_format")
            try:
                scan_info = match.group("scan_info")
            except IndexError:
                scan_info = None
            break
        index += 1
    if original_format:
        metadata["original_format"] = _paren_strip(original_format)
        match_group = 1
        if scan_info:
            metadata["scan_info"] = _paren_strip(scan_info)
            match_group = 0
        _splicey_dicey(data_list, index, match, match_group=match_group)
    else:
        index = 0
    return index


def _pop_value_from_token(
    data_list: list,
    metadata: dict,
    regex: re.Pattern,
    key: str,
    index: int = 0,
):
    """Search token for value, splice and assign to metadata."""
    data = data_list[index]
    match = regex.search(data)
    if match:
        value = _splicey_dicey(data_list, index, match, key)
        metadata[key] = value
    return match


def _parse_item(
    data_list,
    metadata,
    regex,
    key,
    start_index: int = 0,
):
    """Parse a value from the data list into metadata and alter the data list."""
    index = start_index
    dl_len = end_index = len(data_list)
    if index >= end_index:
        index = 0
    while index < end_index:
        match = _pop_value_from_token(data_list, metadata, regex, key, index)
        if match:
            break
        index += 1
        if index > dl_len and start_index > 0:
            index = 0
            end_index = start_index
    return index


def _pop_issue_from_text_fields(data_list, metadata, index):
    """Search issue from ends of text fields."""
    if "issue" not in metadata:
        _pop_value_from_token(data_list, metadata, ISSUE_END_RE, "issue", index=index)
    if "issue" not in metadata:
        _pop_value_from_token(data_list, metadata, ISSUE_BEGIN_RE, "issue", index=index)
    return data_list.pop(index)


def _assign_remaining_groups(data_list, metadata):
    """Assign series and title."""
    index = 0
    for key in _REMAINING_GROUP_KEYS:
        try:
            data = data_list[index]
        except (IndexError, TypeError):
            break
        match = REMAINING_GROUP_RE.search(data) if data else None
        if match:
            value = _pop_issue_from_text_fields(data_list, metadata, index)
            value = _paren_strip(value)
            if value:
                metadata[key] = value
        else:
            index += 1


def _pickup_issue(remainders, metadata):
    """Get issue from remaining tokens or anywhere in a pinch."""
    if "issue" in metadata:
        return
    _parse_item(remainders, metadata, ISSUE_TOKEN_RE, "issue")
    if "issue" in metadata:
        return
    _parse_item(remainders, metadata, ISSUE_ANYWHERE_RE, "issue")


def comicfn2dict(path):
    """Parse the filename with a hierarchy of regexes."""
    metadata = {}
    data_list = _get_data_list(path, metadata)

    # Parse paren tokens
    _parse_item(data_list, metadata, ISSUE_COUNT_RE, "issue_count")
    _parse_item(data_list, metadata, YEAR_TOKEN_RE, "year")
    of_index = _parse_original_format_and_scan_info(data_list, metadata)
    if "original_format" not in metadata:
        of_index = _parse_item(
            data_list, metadata, ORIGINAL_FORMAT_RE, "original_format"
        )
    if "scan_info" not in metadata:
        # Start searching for scan_info after original format.
        _parse_item(
            data_list,
            metadata,
            SCAN_INFO_RE,
            "scan_info",
            start_index=of_index + 1,
        )

    # Parse regular tokens
    _parse_item(data_list, metadata, VOLUME_RE, "volume")
    _parse_item(data_list, metadata, ISSUE_NUMBER_RE, "issue")

    # Pickup year if not gotten.
    if "year" not in metadata:
        _parse_item(data_list, metadata, YEAR_BEGIN_RE, "year")
    if "year" not in metadata:
        _parse_item(data_list, metadata, YEAR_END_RE, "year")

    # Pickup issue if it's a standalone token
    if "issue" not in metadata:
        _parse_item(data_list, metadata, ISSUE_TOKEN_RE, "issue")

    # Series and Title. Also looks for issue.
    _assign_remaining_groups(data_list, metadata)

    # Final try for issue number.
    _pickup_issue(data_list, metadata)

    # Add Remainders
    if data_list:
        metadata["remainders"] = tuple(data_list)

    return metadata



"""Unparse comic filenames."""
from typing import Callable


def issue_formatter(issue):
    """Formatter to zero pad issues."""
    i = 0
    issue = issue.lstrip("0")
    for c in issue:
        if not c.isdigit():
            break
        i += 1
    pad = 3 + len(issue) - i
    return "#{:0>" + str(pad) + "}"


_PAREN_FMT = "({})"
_FILENAME_FORMAT_TAGS = (
    ("series", "{}"),
    ("volume", "v{}"),
    ("issue", issue_formatter),
    ("issue_count", "(of {:03})"),
    ("year", _PAREN_FMT),
    ("title", "{}"),
    ("original_format", _PAREN_FMT),
    ("scan_info", _PAREN_FMT),
)
_EMPTY_VALUES = (None, "")


def dict2comicfn(md, ext=True):
    """Get our preferred basename from a metadata dict."""
    if not md:
        return None
    tokens = []
    for tag, fmt in _FILENAME_FORMAT_TAGS:
        val = md.get(tag)
        if val in _EMPTY_VALUES:
            continue
        final_fmt = fmt(val) if isinstance(fmt, Callable) else fmt
        token = final_fmt.format(val).strip()
        if token:
            tokens.append(token)
    fn = " ".join(tokens)
    if remainders := md.get("remainders"):
        remainder = " ".join(remainders)
        fn += f" - {remainder}"
    if ext:
        fn += "." + md.get("ext", "cbz")
    return fn



"""Tests for filename parsing."""
from pprint import pprint
from types import MappingProxyType

import pytest
from deepdiff.diff import DeepDiff

from comicfn2dict import comicfn2dict
from tests.comic_filenames import FNS

ALL_FIELDS = frozenset({"series", "volume", "issue", "issue_count", "year", "ext"})
FIELD_SCHEMA = MappingProxyType({key: None for key in ALL_FIELDS})


@pytest.mark.parametrize("item", FNS.items())
def test_parse_filename(item):
    """Test filename parsing."""
    fn, defined_fields = item
    md = comicfn2dict(fn)
    diff = DeepDiff(defined_fields, md, ignore_order=True)
    print(fn)
    pprint(defined_fields)
    pprint(md)
    pprint(diff)
    assert not diff



#!/usr/bin/env python3
"""Simple cli for comicfn2dict."""
from argparse import ArgumentParser
from pathlib import Path
from pprint import pprint

from comicfn2dict.parse import comicfn2dict


def main():
    """Test parser."""
    description = "Comic book archive read/write tool."
    parser = ArgumentParser(description=description)
    parser.add_argument("path", help="Path of comic filename to parse", type=Path)
    args = parser.parse_args()
    name = args.path.name
    metadata = comicfn2dict(name)
    pprint(metadata)  # noqa:T203


if __name__ == "__main__":
    main()
